"""A small library for providing a pure interface to Flask response building."""

from collections import namedtuple

import flask


class Response(object):
    """A pure representation of all things that go into a request."""

    def as_flask_response(self):
        """Returns a Flask response including flash messages.

        Note that flash messages involve impure effects and cannot be easily rolled back.
        """
        raise NotImplementedError()  # pragma: no cover


class TemplateResponse(
    Response,
    namedtuple('TemplateResponse', 'template template_args flash_messages')
):
    """A response based on a template."""

    def render_flask_template(self):
        return flask.render_template(self.template, **self.template_args)

    def as_flask_response(self):
        for message, category in self.flash_messages:
            flask.flash(message, category)
        return self.render_flask_template()


class Flash(namedtuple('Flash', 'message category')):
    pass


class RedirectResponse(
    Response,
    namedtuple('RedirectResponse', 'url flash_messages')
):
    """A redirect response."""

    def as_flask_response(self):
        for message, category in self.flash_messages:
            flask.flash(message, category)
        return flask.redirect(self.url)


class Responder(object):
    """Totally encapsulates inputs and logic to form a response."""
    default_next_endpoint = None
    template = None

    def __init__(self, query_args=None, form_kwargs=None, default_template_args=None):
        self.query_args = query_args or {}
        self.default_template_args = default_template_args or {}
        self.form_kwargs = form_kwargs or {}

    def get_next_url(self):
        return self.query_args.get('next', self.url_for(self.default_next_endpoint))

    def render_with(self, template_args=None, flash_messages=()):
        args = self.default_template_args.copy()
        args.update(template_args or {})

        return TemplateResponse(self.template, args, flash_messages)

    def url_for(self, *args, **kwargs):
        """Allows responder to be isolated from app configuration."""
        return flask.url_for(*args, **kwargs)  # pragma: no cover


class LoggedInResponder(Responder):
    """A `Responder` that relies on a known user."""
    def __init__(self, user, *args, **kwargs):
        self.user = user
        super(LoggedInResponder, self).__init__(*args, **kwargs)



from datetime import date
import binascii

import flask
from keg_elements.crypto import (
    constant_time_compare,
    salted_hmac,
)
from keg_elements.http import (
    base36_to_int,
    int_to_base36,
)
from keg_elements.encoding import force_bytes


class PasswordResetTokenGenerator:
    """Generate secure password reset tokens

    This code was largely borrowed from Django and adapted for use in the Keg
    ecosystem.
    """

    key_salt = "keg_login.libs.tokens.PasswordResetTokenGenerator"

    @property
    def timeout_days(self):
        """Number of days before the reset token will timeout"""
        return flask.current_app.config.get('PASSWORD_RESET_TIMEOUT_DAYS', 1)

    @property
    def _secret(self):
        return flask.current_app.config['PASSWORD_RESET_SECRET']

    def make_token(self, user):
        """Return a token that can be used once to do a password reset for the given user."""
        return self._make_token_with_timestamp(user, self._num_days(self._today()))

    def check_token(self, user, token):
        """Check that a password reset token is correct for a given user."""
        if not (user and token):
            return False

        # Parse the token
        try:
            ts_b36, hash = token.split("-")
        except ValueError:
            return False

        try:
            ts = base36_to_int(ts_b36)
        except ValueError:
            return False

        # Check that the timestamp/uid has not been tampered with in a way that
        # isn't seceptiable to timing attacks
        if not constant_time_compare(
            # Regenerate the toke with the user and the timestamp when we
            # created the token in the first place. This should generate an
            # identitcal token from when we created it last time. If anything
            # changed in the _make_hash_function, this will fail.
            force_bytes(self._make_token_with_timestamp(user, ts)),

            # Compare that against the token we received from the URL which, if
            # untampered is what was sent to the user.
            force_bytes(token)
        ):
            return False

        # Check the timestamp is within limit
        if (self._num_days(self._today()) - ts) > self.timeout_days:
            return False

        return True

    def _make_token_with_timestamp(self, user, timestamp):
        # timestamp is number of days since 2001-1-1.  Converted to
        # base 36, this gives us a 3 digit string until about 2121
        ts_b36 = int_to_base36(timestamp)

        # By hashing on the internal state of the user and using state that is sure to change we
        # produce a hash that will be invalid as soon as it is used.
        hash = salted_hmac(
            self.key_salt,
            self._make_hash_value(user, timestamp),
            self._secret
        )

        # This generates a really long hex value  based on the hash from the
        # HMAC... we really don't need all of it, so we trim it down to keep the
        # url shorter.
        prepared_hash = binascii.hexlify(hash).decode()[::2]

        return "%s-%s" % (ts_b36, prepared_hash)

    def _make_hash_value(self, user, timestamp):
        login_timestamp = ('' if user.last_login is None
                           else user.last_login.replace(microsecond=0, tzinfo=None))

        return str(user.id) + str(user.password) + str(login_timestamp) + str(timestamp)

    def _num_days(self, dt):
        return (dt - date(2001, 1, 1)).days

    def _today(self):
        # Used for mocking in tests
        return date.today()



import jinja2


__all__ = [
    'KegLogin'
]


class KegLogin(object):
    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        loader = jinja2.ChoiceLoader([
            app.jinja_loader,
            jinja2.PackageLoader('keg_login', 'templates')
        ])
        app.jinja_loader = loader



from __future__ import absolute_import

import flask
import flask_login
import keg.web
import wrapt
from wtforms.validators import ValidationError

from . import forms
from .lib import (
    Flash,
    LoggedInResponder,
    RedirectResponse,
    Responder,
)

__all__ = [
    'ChangePassword',
    'ForgotPassword',
    'Login',
    'Logout',
    'ResetPassword',
]


def requires_login(current_user=flask_login.current_user):
    @wrapt.decorator
    def decorator(fn, instance, args, kwargs):
        return (fn(*args, **kwargs) if current_user and current_user.is_authenticated
                else flask.abort(401))

    return decorator


def to_wtforms_validator(validator):
    def wtforms_validator(form, field):
        return validator(field.data)
    return wtforms_validator


class KegLoginView(keg.web.BaseView):
    url = None

    class Responder(Responder):
        pass

    def get_responder(self):
        return self.Responder(
            query_args=flask.request.args,
            default_template_args=self.base_template_args()
        )


class CanLogin(object):
    """Mixin to capture the login effect. Defaults to Flask-Login's behavior."""
    def login_user(self, user, remember, *args, **kwargs):
        return flask_login.login_user(user, remember, *args, **kwargs)  # pragma: no cover


class CanLogout(object):
    """Mixin to capture the logout effect. Defaults to Flask-Login's behavior."""
    def logout_user(self):
        flask_login.logout_user()  # pragma: no cover


class NeedsCurrentUser(object):
    """Mixin to capture the effect of looking at the current user."""
    def get_current_user(self):
        return flask_login.current_user  # pragma: no cover


class ChangePassword(KegLoginView, NeedsCurrentUser):
    url = '/change-password'

    class Responder(LoggedInResponder):
        template = 'keg-login/change-password.html'

        def verify_old_password(self, password):
            return self.user.verify_password(password)

        def new_password_validators(self):
            pass  # pragma: no cover

        def make_form(self):
            _validate_old_password = forms.bool_validator(
                lambda data: self.verify_old_password(data),
                u'Old password is not correct'
            )

            form_cls = forms.make_change_password_form(
                [_validate_old_password],
                list(map(to_wtforms_validator, self.new_password_validators()))
            )
            return form_cls(next=self.get_next_url(), **(self.form_kwargs or {}))

        def update_password(self, user, password):
            user.set_password(password)

        def handle_successful_reset(self):
            """Returns a list of `Flash` messages."""
            return [Flash(u'Your password has been changed successfully.', 'success')]

        def get(self):
            return self.render_with({'form': self.make_form()})

        def post(self):
            form = self.make_form()

            if form.validate():
                self.update_password(self.user, form.new_password.data)
                flashes = self.handle_successful_reset()
                return RedirectResponse(form.next.data, flashes)

            return self.render_with({'form': form})

    def get_responder(self):
        current_user = self.get_current_user()
        return requires_login(current_user)(
            lambda: self.Responder(
                user=current_user,
                query_args=flask.request.args,
                default_template_args=self.base_template_args()
            )
        )()

    def get(self):
        return self.get_responder().get().as_flask_response()

    def post(self):
        return self.get_responder().post().as_flask_response()


class ForgotPassword(KegLoginView):
    url = '/forgot-password'

    class Responder(Responder):
        template = 'keg-login/forgot-password.html'

        def make_form(self):
            form_cls = forms.make_forgot_password_form()
            return form_cls(next=self.get_next_url(), **(self.form_kwargs or {}))

        def request_password_reset(self, email):
            """Do actions necessary to request a new password for the given email.

            :returns: `Flash` messages.
            """
            raise NotImplementedError()  # pragma: no cover

        def get(self):
            return self.render_with({'form': self.make_form()})

        def post(self):
            form = self.make_form()

            if form.validate():
                flashes = self.request_password_reset(form.email.data)
                return RedirectResponse(form.next.data, flashes)

            return self.render_with({'form': form})

    def get(self):
        return self.get_responder().get().as_flask_response()

    def post(self):
        return self.get_responder().post().as_flask_response()


class Login(KegLoginView):
    url = '/login'

    class Responder(Responder, CanLogin):
        template = 'keg-login/login.html'
        enable_remember_me = True

        def get_id_form_field(self, user_password_validator=None):
            """Returns a WTForms field object to represent the ID of a user. For example, this might be
            a username, an email, or both. The field should validate its input."""
            return forms.make_email_login_field(user_password_validator)

        def get_user_by_id(self, id):
            """Returns a user entity based on the id field from the login form or None if no match was
            found."""
            raise NotImplementedError()  # pragma: no cover

        def verify_password(self, user, password):
            return user.verify_password(password)

        def make_form(self):
            def validate_password(form):
                user = self.get_user_by_id(form.id.data)
                if not user or not self.verify_password(user, form.password.data):
                    raise ValidationError(u'Invalid user or password')

            form_cls = forms.make_login_form(
                validate_password,
                self.get_id_form_field(),
                enable_remember_me=self.enable_remember_me
            )
            return form_cls(next=self.get_next_url(), **(self.form_kwargs or {}))

        def check_user_blocked_from_login_attempt(self, user):
            """Returns `Flash` messages IFF the given user is not even permitted to attempt a login.
            This check happens if a user can be found for the given identity."""
            return []  # pragma: no cover

        def check_user_blocked_from_login(self, user):
            """Returns `Flash` messages IFF the given user is blocked from logging in even after
            providing proper credentials."""
            return []  # pragma: no cover

        def handle_login_success(self, user):
            """Does some stuff when a user enters the proper credentials and is not blocked from
            login.

            :returns: `Flash` messages.
            """
            return [Flash('Welcome!', 'success')]  # pragma: no cover

        def handle_login_failure(self, user):
            """Does some stuff when a valid user enters a wrong password. E.g. log messages, etc."""
            pass  # pragma: no cover

        def get_current_user(self):
            return flask_login.current_user  # pragma: no cover

        def get(self):
            return self.render_with({'form': self.make_form()})

        def post(self):
            form = self.make_form()
            user = self.get_user_by_id(form.id.data)

            attempt_blocked_flashes = (self.check_user_blocked_from_login_attempt(user)
                                       if user else [])
            if attempt_blocked_flashes:
                return self.render_with({'form': form}, attempt_blocked_flashes)

            if form.validate():
                login_blocked_flashes = self.check_user_blocked_from_login(user)
                if login_blocked_flashes:
                    return self.render_with({'form': form}, login_blocked_flashes)

                flashes = self.handle_login_success(user)
                self.login_user(user, remember=self.enable_remember_me and form.remember_me.data)
                return RedirectResponse(form.next.data, flashes)

            if user:
                self.handle_login_failure(user)

            return self.render_with({'form': form}, ())

    def get(self):
        responder = self.get_responder()
        current_user = responder.get_current_user()
        return (
            RedirectResponse(responder.get_next_url(), [])  # Skip login form when logged in.
            if current_user and current_user.is_authenticated
            else responder.get()
        ).as_flask_response()

    def post(self):
        return self.get_responder().post().as_flask_response()


class Logout(KegLoginView, NeedsCurrentUser):
    url = '/logout'

    class Responder(LoggedInResponder, CanLogout):
        def get(self):
            self.logout_user()
            return RedirectResponse(
                self.get_next_url(),
                [Flash(u'You have logged out successfully.', 'success')]
            )

    def get_responder(self):
        return self.Responder(user=self.get_current_user(), query_args=flask.request.args)

    def get(self):
        return self.get_responder().get().as_flask_response()


class ResetPassword(KegLoginView, NeedsCurrentUser, CanLogout):
    url = '/reset-password/<token>'

    class Responder(Responder):
        template = 'keg-login/reset-password.html'

        def __init__(self, token, *args, **kwargs):
            self.token = token
            super(ResetPassword.Responder, self).__init__(*args, **kwargs)

        class InvalidResetTokenError(Exception):
            default_error_message = u'That reset token is invalid.'

            def __init__(self, message=None):
                self.message = message or self.default_error_message
                super(ResetPassword.Responder.InvalidResetTokenError, self).__init__(
                    self.message)

        class ExpiredResetTokenError(InvalidResetTokenError):
            default_error_message = u'That reset token has expired.'

        def process_reset_token(self, reset_token):
            """Returns a user object and any additional data for the given reset token.

            :param reset_token: is the token to use for looking up a user.
            :returns: a 2-tuple like (user, token_data) where
                        * user is the user entity for the reset token or
                          None to imply `self.InvalidResetTokenError`.
                        * token_data is any other data to remember about the token.

            :raises: `self.InvalidResetTokenError`

            `InvalidResetTokenErrors` are flashed to the page for the user to see.
            """
            raise NotImplementedError()  # pragma: no cover

        def update_password(self, user, token_data, new_password):
            user.set_password(new_password)

        def handle_successful_reset(self, user, token_data, reset_token, new_password):
            """Does any action necessary to mark a reset token as being used."""
            self.update_password(user, token_data, new_password)
            return RedirectResponse(
                self.get_next_url(),
                [Flash(u'Your password has been reset successfully.', 'success')]
            )

        def new_password_validators(self, user):
            return []  # pragma: no cover

        def make_form(self, user, token_data):
            password_validators = list(
                map(to_wtforms_validator, self.new_password_validators(user))
            )
            form_cls = forms.make_reset_password_form(password_validators)
            return form_cls(next=self.get_next_url(), **(self.form_kwargs or {}))

        def respond(self, is_post):
            user = None
            try:
                user, token_data = self.process_reset_token(self.token)
                if not user:
                    raise self.InvalidResetTokenError()
            except self.InvalidResetTokenError as e:
                return RedirectResponse(self.get_next_url(), [Flash(e.message, 'error')])

            form = self.make_form(user, token_data)

            if is_post and form.validate():
                return self.handle_successful_reset(
                    user,
                    token_data,
                    self.token,
                    form.new_password.data
                )

            return self.render_with({'form': form})

        def get(self):
            return self.respond(False)

        def post(self):
            return self.respond(True)

    def get_responder(self, token):
        return self.Responder(
            token=token,
            query_args=flask.request.args,
            default_template_args=self.base_template_args()
        )

    def get(self, token):
        current_user = self.get_current_user()
        if current_user and current_user.is_authenticated:
            self.logout_user()
        return self.get_responder(token).get().as_flask_response()

    def post(self, token):
        return self.get_responder(token).post().as_flask_response()



from pathlib import Path

from blazeutils import randchars
import flask
from pyquery import PyQuery
from wtforms import validators

from keg_login import views
from keg_login.lib import TemplateResponse, RedirectResponse, Flash

template_dir = str(Path(__file__).parent.parent / 'templates')

bp = flask.Blueprint('keg_login', __name__, template_folder=template_dir)
app = flask.Flask(__name__, template_folder=template_dir)

app.config['WTF_CSRF_ENABLED'] = False
app.secret_key = randchars()


class User(object):
    def __init__(self, email, password, is_active=True, is_authenticated=False):
        self.email = email
        self.password = password
        self.is_active = is_active
        self.is_authenticated = is_authenticated

    def verify_password(self, pw):
        return pw == self.password

    def set_password(self, pw):
        self.password = pw


class ResponderTestBase(object):
    def make_user(self, email=None, password=None):
        return User(email or randchars(), password or randchars())

    def make_responder(self, *args, **kwargs):
        return self.Responder(*args, **kwargs)

    def test_next_url(self):
        responder = self.make_responder()
        assert responder.get_next_url() == '/'

        responder = self.make_responder(query_args={'next': 'foo/bar.html'})
        assert responder.get_next_url() == 'foo/bar.html'


class TestChangePasswordResponder(ResponderTestBase):
    class Responder(views.ChangePassword.Responder):
        default_next_endpoint = ''

        def new_password_validators(self):
            return []

        def url_for(self, *args, **kwargs):
            return '/'

    def test_change_password_invalid_old_password(self):
        user = self.make_user(password=None)
        user.set_password('A-password')

        with app.test_request_context():
            responder = self.Responder(user, form_kwargs={
                'old_password': u'B-password',
                'new_password': u'C-password',
                'retype_password': u'C-password',
            })

            response = responder.post()

        assert type(response) == TemplateResponse
        assert response.template_args['form'].old_password.errors == ['Old password is not correct']
        assert response.flash_messages == ()
        assert user.verify_password('A-password')

    def test_change_password_successful(self):
        user = self.make_user(password=None)
        user.set_password('A-password')

        with app.test_request_context():
            responder = self.Responder(user, form_kwargs={
                'old_password': u'A-password',
                'new_password': u'B-password',
                'retype_password': u'B-password',
            })

            response = responder.post()

        assert type(response) == RedirectResponse
        assert response.flash_messages == [
            Flash('Your password has been changed successfully.', 'success')]
        assert response.url == '/'
        assert user.verify_password('B-password')

    def test_get(self):
        with app.test_request_context():
            responder = self.Responder(None)
            response = responder.get()
        assert type(response) == TemplateResponse
        assert response.template_args['form']

    def make_responder(self, *args, **kwargs):
        return self.Responder(None, *args, **kwargs)

current_user = None


class TestChangePasswordView(object):
    class ChangePassword(views.ChangePassword):
        blueprint = bp

        def get_current_user(self):
            return current_user

        def base_template_args(self):
            return {}

        class Responder(views.ChangePassword.Responder):
            def url_for(self, *args, **kwargs):
                return '/'

            def new_password_validators(self):
                return []

    """Tests including full request cycle and all effects."""

    def test_change_password_view_requires_being_logged_in(self):
        global current_user
        current_user = User('user@example.com', 'passW0rd')
        app.register_blueprint(bp)

        test_app = app.test_client()
        response = test_app.get('/change-password')
        assert response.status_code == 401

        response = test_app.post('/change-password')
        assert response.status_code == 401

    def test_change_password_view_get(self):
        global current_user
        current_user = User('user@example.com', 'passW0rd', is_authenticated=True)
        test_app = app.test_client()
        response = test_app.get('/change-password')
        assert response.status_code == 200

    def test_change_password_view_post(self):
        global current_user
        current_user = User('user@example.com', 'passW0rd', is_authenticated=True)
        test_app = app.test_client()
        response = test_app.post('/change-password', data={
            'old_password': 'passW0rd',
            'new_password': 'new-pass1',
            'retype_password': 'new-pass1'
        })
        assert response.status_code == 302
        assert current_user.password == 'new-pass1'


class TestForgotPasswordResponder(ResponderTestBase):
    """Tests isolating [most] dependencies and effects."""

    class Responder(views.ForgotPassword.Responder):
        def __init__(self, *args, **kwargs):
            self.resets_requested = []
            views.ForgotPassword.Responder.__init__(self, *args, **kwargs)

        def url_for(self, *args, **kwargs):
            return '/'

        def request_password_reset(self, email):
            self.resets_requested.append(email)
            return [Flash('Reset requested', 'success')]

    def test_forgot_password_get(self):
        with app.test_request_context():
            response = self.Responder().get()
        assert type(response) == TemplateResponse
        assert response.template_args['form']

    def test_forgot_password_invalid_email(self):
        with app.test_request_context():
            responder = self.Responder(form_kwargs={'email': 'blah'})
            response = responder.post()
        assert type(response) == TemplateResponse
        assert response.template_args['form'].email.errors == ['Invalid email address.']

    def test_forgot_password_success(self):
        with app.test_request_context():
            responder = self.Responder(None, form_kwargs={'email': 'foo@bar.com'})
            response = responder.post()

        assert type(response) == RedirectResponse
        assert response.url == '/'
        assert response.flash_messages == [Flash('Reset requested', 'success')]
        assert responder.resets_requested == ['foo@bar.com']


class TestForgotPasswordView(object):
    class ForgotPassword(views.ForgotPassword):
        blueprint = bp

        def base_template_args(self):
            return {}

        class Responder(views.ForgotPassword.Responder):
            def url_for(self, *args, **kwargs):
                return '/'

            def request_password_reset(self, email):
                return [Flash('', 'success')]

    def setup_method(self, _):
        global current_user
        current_user = None

    """Tests including full request cycle and all effects."""

    def test_forgot_password_view_does_not_require_being_logged_in(self):
        global current_user
        current_user = User('user@example.com', 'passW0rd')
        app.register_blueprint(bp)

        test_app = app.test_client()
        response = test_app.get('/forgot-password')
        assert response.status_code == 200

        response = test_app.post('/forgot-password', data={})
        assert response.status_code == 200

    def test_forgot_password_view_get(self):
        test_app = app.test_client()
        response = test_app.get('/forgot-password')
        assert response.status_code == 200

    def test_forgot_password_view_post(self):
        global current_user
        current_user = User('user@example.com', 'passW0rd', is_authenticated=True)
        test_app = app.test_client()
        response = test_app.post('/forgot-password', data={'email': 'foo@bar.com'})
        assert response.status_code == 302


user_map = {}


class TestLoginResponder(ResponderTestBase):
    def make_user(self, email=None, password=None, is_active=True):
        email = email or '{}@{}.com'.format(randchars(), randchars())
        user = User(email, password or randchars(), is_active)
        user_map[user.email] = user
        return user

    class Responder(views.Login.Responder):
        def __init__(self, *args, **kwargs):
            self.login_effects = []
            views.Login.Responder.__init__(self, *args, **kwargs)

        def get_user_by_id(self, id):
            return user_map.get(id)

        def login_user(self, user, remember, *args, **kwargs):
            self.login_effects.append((user, remember))

        def url_for(self, *args, **kwargs):
            return '/'

        def check_user_blocked_from_login_attempt(self, user):
            if user.email == 'blocked_user@example.com':
                return [Flash('Blocked User', 'error')]
            return []

        def check_user_blocked_from_login(self, user):
            if not user.is_active:
                return [Flash('Inactive User', 'error')]
            return []

    def assert_login_failure(self, user, message):
        with app.test_request_context():
            responder = self.Responder(form_kwargs={
                'id': user.email,
                'password': user.password
            })
            response = responder.post()

        assert type(response) == TemplateResponse
        assert response.flash_messages == [Flash(message, 'error')]
        assert responder.login_effects == []

    def test_login_get(self):
        with app.test_request_context():
            response = self.Responder().get()
        assert type(response) == TemplateResponse
        assert response.template_args['form']

    def test_login_blocked_user(self):
        self.assert_login_failure(
            self.make_user(email='blocked_user@example.com', password='123'),
            'Blocked User'
        )

    def test_login_inactive(self):
        self.assert_login_failure(
            self.make_user(is_active=False),
            'Inactive User'
        )

    def test_login_bad_password(self):
        user = self.make_user()
        with app.test_request_context():
            responder = self.Responder(form_kwargs={
                'id': user.email,
                'password': 'bad-password'
            })

            response = responder.post()

        assert type(response) == TemplateResponse
        assert not response.flash_messages
        assert responder.login_effects == []

    def test_login_success(self):
        with app.test_request_context():
            user = self.make_user(email='user@example.com', password='password123')
            responder = self.Responder(form_kwargs={
                'id': user.email,
                'password': user.password
            })

            response = responder.post()
        assert type(response) == RedirectResponse
        assert response.url == '/'
        assert response.flash_messages == [Flash('Welcome!', 'success')]
        assert responder.login_effects == [(user, False)]

    def test_remember_me_enabled(self):
        with app.test_request_context():
            user = self.make_user()
            responder = self.Responder(form_kwargs={
                'id': user.email,
                'password': user.password,
                'remember_me': True
            })
            response = responder.get()
            assert len(PyQuery(response.as_flask_response())('#remember_me')) == 1

            response = responder.post()
            assert type(response) == RedirectResponse
            assert responder.login_effects == [(user, True)]

    def test_remember_me_disabled(self):
        with app.test_request_context():
            user = self.make_user()
            responder = self.Responder(form_kwargs={
                'id': user.email,
                'password': user.password,
                'remember_me': True
            })
            responder.enable_remember_me = False
            response = responder.get()
            assert len(PyQuery(response.as_flask_response())('[name="remember_me"]')) == 0

            response = responder.post()
            assert type(response) == RedirectResponse
            assert responder.login_effects == [(user, False)]


class TestLoginView(object):
    class Login(views.Login):
        blueprint = bp

        def base_template_args(self):
            return {}

        class Responder(views.Login.Responder):
            def url_for(self, *args, **kwargs):
                return '/'

            def get_current_user(self):
                return current_user

            def login_user(self, user, remember, *args, **kwargs):
                global current_user
                current_user = user

            def get_user_by_id(self, id):
                return user_map.get(id)

            def check_user_blocked_from_login(self, user):
                if not user.is_active:
                    return [Flash('Inactive User', 'error')]
                return []

    def setup_method(self, _):
        global current_user
        current_user = None

    """Tests including full request cycle and all effects."""

    def test_login_view_does_not_require_being_logged_in(self):
        global current_user
        current_user = User('user@example.com', 'passW0rd')
        app.register_blueprint(bp)

        test_app = app.test_client()
        response = test_app.get('/login')
        assert response.status_code == 200

        response = test_app.post('/login', data={})
        assert response.status_code == 200

    def test_redirect_already_logged_in(self):
        global current_user
        current_user = User('foo@bar.com', '123', is_authenticated=True)
        app.register_blueprint(bp)

        test_app = app.test_client()
        response = test_app.get('/login')
        assert response.status_code == 302
        assert response.location == 'http://localhost/'

    def test_login_view_get(self):
        test_app = app.test_client()
        response = test_app.get('/login')
        assert response.status_code == 200

    def test_login_view_post(self):
        user = User('foo@bar.com', 'passW0rd')
        user_map['foo@bar.com'] = user

        test_app = app.test_client()
        response = test_app.post('/login', data={'id': 'foo@bar.com', 'password': 'passW0rd'})
        assert response.status_code == 302
        assert current_user == user

    def test_login_view_bad_password(self):
        user = User('foo@bar.com', 'passW0rd')
        user_map['foo@bar.com'] = user

        test_app = app.test_client()
        response = test_app.post('/login', data={'id': 'foo@bar.com', 'password': 'foo'})
        assert response.status_code == 200
        assert current_user is None

    def test_login_view_user_inactive(self):
        user = User('foo@bar.com', 'passW0rd', is_active=False)
        user_map['foo@bar.com'] = user

        test_app = app.test_client()
        response = test_app.post('/login', data={'id': 'foo@bar.com', 'password': 'passW0rd'})
        assert response.status_code == 200
        assert current_user is None


class TestLogoutResponder(ResponderTestBase):
    """Tests isolating [most] dependencies and effects."""

    class Responder(views.Logout.Responder):
        def __init__(self, *args, **kwargs):
            self.logout_effects = 0
            views.Logout.Responder.__init__(self, *args, **kwargs)

        def logout_user(self):
            self.logout_effects += 1

        def url_for(self, *args, **kwargs):
            return '/'

    def make_responder(self, *args, **kwargs):
        return self.Responder(self.make_user(), *args, **kwargs)

    def test_logout(self):
        user = self.make_user()
        responder = self.Responder(user)

        with app.test_request_context():
            response = responder.get()
        assert type(response) == RedirectResponse
        assert response.flash_messages == [Flash('You have logged out successfully.', 'success')]
        assert response.url == '/'
        assert responder.logout_effects == 1

        with app.test_request_context():
            response = responder.get()
        assert type(response) == RedirectResponse
        assert response.flash_messages == [Flash('You have logged out successfully.', 'success')]
        assert response.url == '/'
        assert responder.logout_effects == 2


class TestLogoutView(object):
    class Logout(views.Logout):
        blueprint = bp

        def get_current_user(self):
            return current_user

        class Responder(views.Logout.Responder):
            def url_for(self, *args, **kwargs):
                return '/'

            def logout_user(self):
                global current_user
                current_user = None

    def setup_method(self, _):
        global current_user
        current_user = None

    """Tests including full request cycle and all effects."""

    def test_logout(self):
        global current_user
        current_user = User('foo@example.com', 'password', is_authenticated=True)
        app.register_blueprint(bp)

        test_app = app.test_client()
        response = test_app.get('/logout')
        assert response.status_code == 302
        assert current_user is None


token_map = {}


class TestResetPasswordResponder(ResponderTestBase):
    """Tests isolating [most] dependencies and effects."""

    class Responder(views.ResetPassword.Responder):
        def __init__(self, *args, **kwargs):
            views.ResetPassword.Responder.__init__(self, *args, **kwargs)

        def process_reset_token(self, reset_token):
            return token_map.get(reset_token), reset_token

        def url_for(self, *args, **kwargs):
            return '/'

        def new_password_validators(self, user):
            def valid(d):
                if d.endswith('123'):
                    raise validators.ValidationError('bad password')
            return [valid]

    def make_token(self, token=None, user=None):
        user = user or User('{}@{}.com'.format(randchars(), randchars()), randchars())
        token = token or randchars()
        token_map[token] = user
        return token

    def make_responder(self, *args, **kwargs):
        return self.Responder(self.make_token(), *args, **kwargs)

    def test_invalid_token(self):
        responder = self.Responder(token=b'blah')
        with app.test_request_context():
            for response in [responder.get(), responder.post()]:
                assert type(response) == RedirectResponse
                assert response.url == '/'
                assert response.flash_messages == [Flash('That reset token is invalid.', 'error')]

    def test_reset_get_successful(self):
        responder = self.Responder(token=self.make_token())
        with app.test_request_context():
            response = responder.get()
        assert type(response) == TemplateResponse
        assert response.template_args['form']

    def test_reset_password_successful(self):
        token = self.make_token()
        user = token_map[token]
        responder = self.Responder(token=token, form_kwargs={
            'new_password': 'new-pass',
            'retype_password': 'new-pass'
        })
        with app.test_request_context():
            response = responder.post()
            assert type(response) == RedirectResponse
            assert response.url == '/'
            assert response.flash_messages == [
                Flash('Your password has been reset successfully.', 'success')]
            assert user.password == 'new-pass'

    def test_reset_form_error(self):
        token = self.make_token()
        user = token_map[token]
        responder = self.Responder(token=token, form_kwargs={
            'new_password': 'new-pass',
            'retype_password': 'new-pass1'
        })
        with app.test_request_context():
            response = responder.post()
            assert type(response) == TemplateResponse
            assert not response.flash_messages
            assert user.password != 'new-pass'

    def test_validator_error(self):
        token = self.make_token()
        user = token_map[token]
        responder = self.Responder(token=token, form_kwargs={
            'new_password': 'new-pass123',
            'retype_password': 'new-pass123'
        })
        with app.test_request_context():
            response = responder.post()
            assert type(response) == TemplateResponse
            assert not response.flash_messages
            assert user.password != 'new-pass123'


class TestResetPasswordView(object):
    class ResetPassword(views.ResetPassword):
        blueprint = bp

        def base_template_args(self):
            return {}

        def get_current_user(self):
            return current_user

        def logout_user(self):
            global current_user
            current_user = None

        class Responder(views.ResetPassword.Responder):
            def url_for(self, *args, **kwargs):
                return '/'

            def process_reset_token(self, reset_token):
                return token_map.get(reset_token), reset_token

    def setup_method(self, _):
        global current_user
        current_user = None

    """Tests including full request cycle and all effects."""

    def test_reset_password_post(self):
        user = User('foo@example.com', 'password')
        token = randchars()
        token_map[token] = user

        app.register_blueprint(bp)

        test_app = app.test_client()
        response = test_app.post('/reset-password/{}'.format(token), data={
            'new_password': 'abc123',
            'retype_password': 'abc123'
        })
        assert response.status_code == 302
        assert user.password == 'abc123'

    def test_reset_password_get(self):
        user = User('foo@example.com', 'password')
        token = randchars()
        token_map[token] = user

        app.register_blueprint(bp)

        test_app = app.test_client()
        response = test_app.get('/reset-password/{}'.format(token), )
        assert response.status_code == 200
        assert current_user is None

    def test_reset_password_get_with_logged_in_user(self):
        global current_user
        current_user = User('bar@example.com', 'drowssap', is_authenticated=True)

        user = User('foo@example.com', 'password')
        token = randchars()
        token_map[token] = user

        app.register_blueprint(bp)

        test_app = app.test_client()
        response = test_app.get('/reset-password/{}'.format(token), )
        assert response.status_code == 200
        assert current_user is None

    def test_reset_password_post_bad_token(self):
        app.register_blueprint(bp)

        test_app = app.test_client()
        response = test_app.post('/reset-password/{}'.format(randchars()), data={
            'new_password': 'abc123',
            'retype_password': 'abc123'
        })
        assert response.status_code == 302



import flask
from werkzeug.datastructures import MultiDict
from wtforms import validators, fields

from keg_login.forms import make_change_password_form, make_forgot_password_form, make_login_form, \
    make_email_login_field, make_reset_password_form, bool_validator

app = flask.Flask(__name__)
app.config['WTF_CSRF_ENABLED'] = False


class TestChangePasswordForm(object):
    def form_cls(self, old_pw_validators=None, new_pw_validators=None):
        return make_change_password_form(old_pw_validators or [], new_pw_validators or [])

    def test_ok(self):
        with app.test_request_context():
            data = {
                'old_password': '12345',
                'new_password': 'password',
                'retype_password': 'password'
            }
            form = self.form_cls()(MultiDict(data))
            assert form.validate()

    def test_required(self):
        with app.test_request_context():
            form = self.form_cls()()
            assert not form.validate()
            msg = ['This field is required.']
            assert form.old_password.errors == msg
            assert form.new_password.errors == msg
            assert form.retype_password.errors == []

    def test_retype_validation(self):
        with app.test_request_context():
            data = {
                'old_password': '12345',
                'new_password': 'password',
                'retype_password': 'passw0rd'
            }
            form = self.form_cls()(MultiDict(data))
            assert not form.validate()
            assert form.old_password.errors == []
            assert form.new_password.errors == []
            assert form.retype_password.errors == ['Passwords do not match']

    def test_extra_validators(self):
        with app.test_request_context():
            data = {
                'old_password': '12345',
                'new_password': 'password',
                'retype_password': 'password'
            }
            form = self.form_cls(
                old_pw_validators=[validators.Length(max=3)],
                new_pw_validators=[validators.Email()]
            )(MultiDict(data))

            assert not form.validate()
            assert form.old_password.errors == ['Field cannot be longer than 3 characters.']
            assert form.new_password.errors == ['Invalid email address.']
            assert form.retype_password.errors == []


class TestForgotPasswordForm(object):
    def form_cls(self, email_validator=lambda *a, **kwargs: None):
        return make_forgot_password_form(email_validator)

    def test_ok(self):
        with app.test_request_context():
            data = {'email': 'foo@example.com'}
            form = self.form_cls()(MultiDict(data))
            assert form.validate()

    def test_required(self):
        with app.test_request_context():
            form = self.form_cls()()
            assert not form.validate()
            msg = ['This field is required.']
            assert form.email.errors == msg

    def test_email_validation(self):
        with app.test_request_context():
            form = self.form_cls()(MultiDict({'email': 'foo.bar.com'}))
            assert not form.validate()
            assert form.email.errors == ['Invalid email address.']

    def test_custom_validation(self):
        with app.test_request_context():
            def invalid(*args, **kwargs):
                raise validators.ValidationError('Not Valid')

            form = self.form_cls(invalid)(MultiDict({'email': 'foo@bar.com'}))
            assert not form.validate()
            assert form.email.errors == ['Not Valid']


class TestLoginForm(object):
    def form_cls(self, password_validator=None, id_field=None):
        return make_login_form(
            password_validator or (lambda *args, **kwargs: None),
            id_field or make_email_login_field(lambda *args, **kwargs: None)
        )

    def test_ok(self):
        with app.test_request_context():
            data = {
                'id': 'foo@example.com',
                'password': '12345'
            }
            form = self.form_cls()(MultiDict(data))
            assert form.validate()

    def test_required(self):
        with app.test_request_context():
            form = self.form_cls()()
            assert not form.validate()
            msg = ['This field is required.']
            assert form.id.errors == msg
            assert form.password.errors == msg

    def test_custom_validation(self):
        with app.test_request_context():
            def invalid(*args, **kwargs):
                raise validators.ValidationError('Not Valid')

            form = self.form_cls(invalid)(MultiDict({'id': 'foo@bar.com', 'password': 'Foo'}))
            assert not form.validate()
            assert form.form_errors == ['Not Valid']

    def test_id_field(self):
        with app.test_request_context():
            id = fields.IntegerField('User', validators=[validators.NumberRange(min=10)])
            form = self.form_cls(id_field=id)(MultiDict({'id': '3', 'password': 'Foo'}))
            assert not form.validate()
            assert form.id.errors == ['Number must be at least 10.']

            form = self.form_cls(id_field=id)(MultiDict({'id': '15', 'password': 'Foo'}))
            assert form.validate()

    def test_remember_me_enabled(self):
        with app.test_request_context():
            data = {
                'id': 'foo@example.com',
                'password': '12345',
            }
            form = self.form_cls()(MultiDict(data))
            assert form.remember_me.data is False

            data['remember_me'] = True
            form = self.form_cls()(MultiDict(data))
            assert form.remember_me.data is True

    def test_remember_me_disabled(self):
        form_cls = make_login_form(
            lambda *args, **kwargs: None,
            make_email_login_field(lambda *args, **kwargs: None),
            enable_remember_me=False
        )
        with app.test_request_context():
            data = {
                'id': 'foo@example.com',
                'password': '12345',
                'remember_me': True
            }
            form = form_cls(MultiDict(data))
            assert not hasattr(form, 'remember_me')
            assert form.validate()
            assert 'remember_me' not in form.data


class TestResetPasswordForm(object):
    def form_cls(self, pw_validators=None):
        return make_reset_password_form(pw_validators)

    def test_ok(self):
        with app.test_request_context():
            data = {
                'new_password': 'password',
                'retype_password': 'password'
            }
            form = self.form_cls()(MultiDict(data))
            assert form.validate()

    def test_required(self):
        with app.test_request_context():
            form = self.form_cls()()
            assert not form.validate()
            msg = ['This field is required.']
            assert form.new_password.errors == msg
            assert form.retype_password.errors == []

    def test_retype_validation(self):
        with app.test_request_context():
            data = {
                'new_password': 'password',
                'retype_password': 'passw0rd'
            }
            form = self.form_cls()(MultiDict(data))
            assert not form.validate()
            assert form.new_password.errors == []
            assert form.retype_password.errors == ['Passwords do not match']


class TestBoolValidator(object):
    def form_inst(self, pw_validators=None):
        return make_reset_password_form(pw_validators)(MultiDict({
                'new_password': 'password',
                'retype_password': 'password'
            }))

    def test_ok(self):
        with app.test_request_context():
            form = self.form_inst([bool_validator(lambda x: True, 'Invalid')])
            assert form.validate()

    def test_invalid(self):
        with app.test_request_context():
            form = self.form_inst([bool_validator(lambda x: False, 'Invalid')])
            assert not form.validate()
            assert form.new_password.errors == ['Invalid']



from datetime import datetime as dt, date as d

import pytest

import keg_login.tokens as tokens


@pytest.fixture(scope="module")
def prtg():
    return MockPRTG()


@pytest.fixture(scope="module")
def user():
    return MockUser()


class MockPRTG(tokens.PasswordResetTokenGenerator):
    """Mock out the PRTG with testable values"""

    _testing_today = None
    _testing_secret = None
    _testing_timeout = None

    def setup(self, timeout=None, secret=None, today=None, salt=None):
        self._testing_timeout = timeout
        self._testing_secret = secret
        self._testing_today = today
        self.key_salt = salt or self.key_salt

    @property
    def timeout_days(self):
        return self._testing_timeout or 1

    @property
    def _secret(self):
        return self._testing_secret or 'secret'

    def _today(self):
        return self._testing_today or d(2001, 1, 1)


class MockUser(object):
    def __init__(self, id=1, last_login=dt(2001, 1, 1),
                 password='password'):
        self.id = id
        self.last_login = last_login
        self.password = password


class TestPasswordResetTokenGenerator(object):

    def test_hash_value_changes_with_different_inputs(self, prtg, user):
        ts = dt(2017, 1, 1)
        default = prtg._make_hash_value(user, ts)

        # check for purity
        assert default == prtg._make_hash_value(user, ts)

        # Check different inputs
        assert default != prtg._make_hash_value(user, dt(2001, 1, 2))
        assert default != prtg._make_hash_value(MockUser(id=0), ts)
        assert default != prtg._make_hash_value(MockUser(last_login=dt(2001, 1, 2)), ts)
        assert default != prtg._make_hash_value(MockUser(password='pass'), ts)
        assert default != prtg._make_hash_value(MockUser(last_login=None), ts)

    def test_make_token_prepends_base36_ts_value(self, prtg, user):
        ts = dt(2017, 1, 1)

        ts, token1 = prtg._make_token_with_timestamp(user, 1).split('-')
        assert ts == '1'

        ts, token2 = prtg._make_token_with_timestamp(user, 100).split('-')
        assert ts == '2s'

        assert token1 != token2

    def test_make_token_unique_token(self, prtg, user):
        default = prtg._make_token_with_timestamp(user, 1)

        # Generation on different day should be the same token
        prtg.setup(today=dt(2017, 1, 2))
        assert default == prtg._make_token_with_timestamp(user, 1)

        # Different timestamp should be different token
        assert default != prtg._make_token_with_timestamp(user, 2)

        # different user should be a different token
        assert default != prtg._make_token_with_timestamp(MockUser(id=2), 1)

        # different salt should be a different token
        prtg.setup(salt='salt')
        assert default != prtg._make_token_with_timestamp(user, 1)

        # different secret should be a different token
        prtg.setup(secret='other')
        assert default != prtg._make_token_with_timestamp(user, 1)

    def test_check_token(self, prtg, user):
        default = prtg._make_token_with_timestamp(user, 1)

        assert prtg.check_token(user, default)
        assert not prtg.check_token(user, 'bad')
        assert not prtg.check_token(MockUser(id=2), default)
        assert not prtg.check_token(None, default)
        assert not prtg.check_token(user, None)
        assert not prtg.check_token(None, None)
        assert not prtg.check_token(user, 'a-b')

        # Bad base36
        assert not prtg.check_token(user, 'zzzzzzzzzzzzzz-as')

        # Test token expiration
        prtg.setup(today=d(2017, 1, 3))
        assert not prtg.check_token(user, default)



from keg import Keg
from pyquery import PyQuery
from keg_login.ext import KegLogin
from flask import render_template

from keg_login.forms import make_change_password_form, make_forgot_password_form


class TestExtension(object):
    def get_app(self):
        app = Keg(__name__, template_folder='test_templates')
        app.config['WTF_CSRF_ENABLED'] = False
        app.config['SERVER_NAME'] = 'localhost'
        return app

    def test_loader_override(self):
        app = self.get_app()
        ext = KegLogin()
        ext.init_app(app)
        with app.app_context():
            form = make_forgot_password_form()()
            rendered = render_template('keg-login/forgot-password.html', form=form)
        assert rendered.strip() == 'TEST TEMPLATE'

    def test_loader_passthrough(self):
        app = self.get_app()
        ext = KegLogin()
        ext.init_app(app)

        with app.app_context():
            form = make_change_password_form([], [])()
            rendered = render_template('keg-login/change-password.html', form=form)
        assert PyQuery(rendered)('form input[type="submit"]').val() == 'Change Password'

    def test_import(self):
        app = self.get_app()
        ext = KegLogin()
        ext.init_app(app)

        with app.app_context():
            form = make_change_password_form([], [])()
            rendered = render_template('keg-login/test-include.html', form=form)
        assert PyQuery(rendered)('form input[type="submit"]').val() == 'Change Password'

    def test_init_from_constructor(self):
        app = self.get_app()
        KegLogin(app)

        with app.app_context():
            form = make_forgot_password_form()()
            rendered = render_template('keg-login/forgot-password.html', form=form)
        assert rendered.strip() == 'TEST TEMPLATE'



from keg_login.ext import KegLogin
from keg_login.forms import *  # noqa
from keg_login.views import *  # noqa


__all__ = [
    'KegLogin',
    'make_change_password_form',
    'make_forgot_password_form',
    'make_email_login_field',
    'make_login_form',
    'make_reset_password_form',
    'ChangePassword',
    'ForgotPassword',
    'Login',
    'Logout',
    'ResetPassword',
]



import os.path as osp

from setuptools import setup, find_packages

cdir = osp.abspath(osp.dirname(__file__))
README = open(osp.join(cdir, 'readme.rst')).read()

setup(
    name='KegLogin',
    setup_requires=[
        'setuptools_scm',
    ],
    use_scm_version=True,
    description='Authentication views for Keg',
    author='Level 12',
    author_email='devteam@level12.io',
    url='https://github.com/level12/keg-login',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'Flask-Login',
        'Keg',
        'KegElements'
    ],
    long_description=README,
)



from keg_elements.forms import Form, form_validator
from wtforms.fields import (
    BooleanField,
    HiddenField,
    PasswordField,
    StringField,
)
from wtforms import validators


__all__ = [
    'make_change_password_form',
    'make_forgot_password_form',
    'make_email_login_field',
    'make_login_form',
    'make_reset_password_form',
]


def make_change_password_form(old_password_validators, new_password_validators):
    class ChangePasswordForm(Form):
        next = HiddenField()
        old_password = PasswordField(
            u'Old Password',
            validators=[validators.DataRequired()] + old_password_validators
        )
        new_password = PasswordField(
            u'New Password',
            validators=[validators.DataRequired()] + new_password_validators
        )
        retype_password = PasswordField(u'Retype New Password', validators=[
            validators.EqualTo('new_password', message=u'Passwords do not match')
        ])

    return ChangePasswordForm


def make_forgot_password_form(email_validator=lambda *a, **kw: None):
    class ForgotPasswordForm(Form):
        next = HiddenField()

        email = StringField(u'Your email address', validators=[
            validators.DataRequired(),
            validators.Email(),
            email_validator,
        ])

    return ForgotPasswordForm


def make_email_login_field(email_validator=None):
    vl = [
        validators.DataRequired(),
        validators.Email(),
    ]
    if email_validator is not None:
        vl.append(email_validator)
    return StringField(u'Email', validators=vl)


def make_login_form(password_validator, id_field, enable_remember_me=True):
    class LoginForm(Form):
        next = HiddenField()
        id = id_field

        password = PasswordField('Password', validators=[
            validators.DataRequired(),
        ])

        if enable_remember_me:
            remember_me = BooleanField('Remember me')

        @form_validator
        def _password_validator(self):
            password_validator(self)

    return LoginForm


def make_reset_password_form(password_validators=None):
    class ResetPasswordForm(Form):
        next = HiddenField()
        new_password = PasswordField(
            u'New Password',
            validators=[validators.DataRequired()] + (password_validators or []))
        retype_password = PasswordField(u'Retype New Password', validators=[
            validators.EqualTo('new_password', message=u'Passwords do not match'),
        ])

    return ResetPasswordForm


def bool_validator(pred, error_message):
    """Returns a simple WTForms validator that raises the given error message as a validation error
    IFF the given predicate function returns False on the field's value.
    """
    def validator(form, field):
        if not pred(field.data):
            raise validators.ValidationError(error_message)

    return validator
